@@ -3,10 +3,12 @@ use std::{num::NonZeroU32, ops::Range};
 #[cfg(feature = "trace")]
 use crate::device::trace::Command as TraceCommand;
 use crate::{
+    align_to,
     command::CommandBuffer,
+    get_lowest_common_denom,
     hub::{Global, GlobalIdentityHandlerFactory, HalApi, Token},
-    id::{BufferId, CommandEncoderId, TextureId},
-    memory_init_tracker::{MemoryInitKind, MemoryInitTrackerAction},
+    id::{BufferId, CommandEncoderId, DeviceId, TextureId},
+    init_tracker::MemoryInitKind,
     track::TextureSelector,
 };
 
@@ -23,6 +25,8 @@ pub enum ClearError {
     MissingClearCommandsFeature,
     #[error("command encoder {0:?} is invalid")]
     InvalidCommandEncoder(CommandEncoderId),
+    #[error("device {0:?} is invalid")]
+    InvalidDevice(DeviceId),
     #[error("buffer {0:?} is invalid or destroyed")]
     InvalidBuffer(BufferId),
     #[error("texture {0:?} is invalid or destroyed")]
@@ -44,6 +48,10 @@ pub enum ClearError {
         texture_format: wgt::TextureFormat,
         subresource_range_aspects: TextureAspect,
     },
+    #[error("Depth/Stencil formats are not supported for clearing")]
+    DepthStencilFormatNotSupported,
+    #[error("Multisampled textures are not supported for clearing")]
+    MultisampledTextureUnsupported,
     #[error("image subresource level range is outside of the texture's level range. texture range is {texture_level_range:?},  \
 whereas subesource range specified start {subresource_base_mip_level} and count {subresource_mip_level_count:?}")]
     InvalidTextureLevelRange {
@@ -82,10 +90,6 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
             list.push(TraceCommand::ClearBuffer { dst, offset, size });
         }
 
-        if !cmd_buf.support_fill_buffer_texture {
-            return Err(ClearError::MissingClearCommandsFeature);
-        }
-
         let (dst_buffer, dst_pending) = cmd_buf
             .trackers
             .buffers
@@ -127,37 +131,34 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         }
 
         // Mark dest as initialized.
-        cmd_buf.buffer_memory_init_actions.extend(
-            dst_buffer
-                .initialization_status
-                .check(offset..end)
-                .map(|range| MemoryInitTrackerAction {
-                    id: dst,
-                    range,
-                    kind: MemoryInitKind::ImplicitlyInitialized,
-                }),
-        );
-
+        cmd_buf
+            .buffer_memory_init_actions
+            .extend(dst_buffer.initialization_status.create_action(
+                dst,
+                offset..end,
+                MemoryInitKind::ImplicitlyInitialized,
+            ));
         // actual hal barrier & operation
         let dst_barrier = dst_pending.map(|pending| pending.into_hal(dst_buffer));
         let cmd_buf_raw = cmd_buf.encoder.open();
         unsafe {
             cmd_buf_raw.transition_buffers(dst_barrier);
-            cmd_buf_raw.fill_buffer(dst_raw, offset..end, 0);
+            cmd_buf_raw.clear_buffer(dst_raw, offset..end);
         }
         Ok(())
     }
 
-    pub fn command_encoder_clear_image<A: HalApi>(
+    pub fn command_encoder_clear_texture<A: HalApi>(
         &self,
         command_encoder_id: CommandEncoderId,
         dst: TextureId,
         subresource_range: &ImageSubresourceRange,
     ) -> Result<(), ClearError> {
-        profiling::scope!("CommandEncoder::clear_image");
+        profiling::scope!("CommandEncoder::clear_texture");
 
         let hub = A::hub(self);
         let mut token = Token::root();
+        let (device_guard, mut token) = hub.devices.write(&mut token);
         let (mut cmd_buf_guard, mut token) = hub.command_buffers.write(&mut token);
         let cmd_buf = CommandBuffer::get_encoder_mut(&mut *cmd_buf_guard, command_encoder_id)
             .map_err(|_| ClearError::InvalidCommandEncoder(command_encoder_id))?;
@@ -166,13 +167,13 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
 
         #[cfg(feature = "trace")]
         if let Some(ref mut list) = cmd_buf.commands {
-            list.push(TraceCommand::ClearImage {
+            list.push(TraceCommand::ClearTexture {
                 dst,
                 subresource_range: subresource_range.clone(),
             });
         }
 
-        if !cmd_buf.support_fill_buffer_texture {
+        if !cmd_buf.support_clear_buffer_texture {
             return Err(ClearError::MissingClearCommandsFeature);
         }
 
@@ -189,6 +190,15 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                 subresource_range_aspects: subresource_range.aspect,
             });
         };
+
+        // Check if texture is supported for clearing
+        if dst_texture.desc.format.describe().sample_type == wgt::TextureSampleType::Depth {
+            return Err(ClearError::DepthStencilFormatNotSupported);
+        }
+        if dst_texture.desc.sample_count > 1 {
+            return Err(ClearError::MultisampledTextureUnsupported);
+        }
+
         // Check if subresource level range is valid
         let subresource_level_end = match subresource_range.mip_level_count {
             Some(count) => subresource_range.base_mip_level + count.get(),
@@ -232,7 +242,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                 hal::TextureUses::COPY_DST,
             )
             .map_err(ClearError::InvalidTexture)?;
-        let _dst_raw = dst_texture
+        let dst_raw = dst_texture
             .inner
             .as_raw()
             .ok_or(ClearError::InvalidTexture(dst))?;
@@ -242,27 +252,101 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
 
         // actual hal barrier & operation
         let dst_barrier = dst_pending.map(|pending| pending.into_hal(dst_texture));
-        let cmd_buf_raw = cmd_buf.encoder.open();
+        let encoder = cmd_buf.encoder.open();
+        let device = &device_guard[cmd_buf.device_id.value];
+
+        let mut zero_buffer_copy_regions = Vec::new();
+        collect_zero_buffer_copies_for_clear_texture(
+            &dst_texture.desc,
+            device.alignments.buffer_copy_pitch.get() as u32,
+            subresource_range.base_mip_level..subresource_level_end,
+            subresource_range.base_array_layer..subresource_layer_end,
+            &mut zero_buffer_copy_regions,
+        );
         unsafe {
-            cmd_buf_raw.transition_textures(dst_barrier);
-            /*TODO: image clears
-            cmd_buf_raw.clear_image(
-                dst_raw,
-                hal::image::Layout::TransferDstOptimal,
-                hal::command::ClearValue {
-                    color: hal::command::ClearColor {
-                        float32: conv::map_color_f32(&wgt::Color::TRANSPARENT),
-                    },
-                },
-                std::iter::once(hal::image::SubresourceRange {
-                    aspects,
-                    level_start: subresource_range.base_mip_level as u8,
-                    level_count: subresource_range.mip_level_count.map(|c| c.get() as u8),
-                    layer_start: subresource_range.base_array_layer as u16,
-                    layer_count: subresource_range.array_layer_count.map(|c| c.get() as u16),
-                }),
-            );*/
+            encoder.transition_textures(dst_barrier);
+            if !zero_buffer_copy_regions.is_empty() {
+                encoder.copy_buffer_to_texture(
+                    &device.zero_buffer,
+                    dst_raw,
+                    zero_buffer_copy_regions.into_iter(),
+                );
+            }
         }
         Ok(())
     }
 }
+
+pub(crate) fn collect_zero_buffer_copies_for_clear_texture(
+    texture_desc: &wgt::TextureDescriptor<()>,
+    buffer_copy_pitch: u32,
+    mip_range: Range<u32>,
+    layer_range: Range<u32>,
+    out_copy_regions: &mut Vec<hal::BufferTextureCopy>, // TODO: Something better than Vec
+) {
+    let format_desc = texture_desc.format.describe();
+
+    let bytes_per_row_alignment =
+        get_lowest_common_denom(buffer_copy_pitch, format_desc.block_size as u32);
+
+    for mip_level in mip_range {
+        let mut mip_size = texture_desc.mip_level_size(mip_level).unwrap();
+        // Round to multiple of block size
+        mip_size.width = align_to(mip_size.width, format_desc.block_dimensions.0 as u32);
+        mip_size.height = align_to(mip_size.height, format_desc.block_dimensions.1 as u32);
+
+        let bytes_per_row = align_to(
+            mip_size.width / format_desc.block_dimensions.0 as u32 * format_desc.block_size as u32,
+            bytes_per_row_alignment,
+        );
+
+        let max_rows_per_copy = crate::device::ZERO_BUFFER_SIZE as u32 / bytes_per_row;
+        // round down to a multiple of rows needed by the texture format
+        let max_rows_per_copy = max_rows_per_copy / format_desc.block_dimensions.1 as u32
+            * format_desc.block_dimensions.1 as u32;
+        assert!(max_rows_per_copy > 0, "Zero buffer size is too small to fill a single row of a texture with format {:?} and desc {:?}",
+                        texture_desc.format, texture_desc.size);
+
+        let z_range = 0..(if texture_desc.dimension == wgt::TextureDimension::D3 {
+            mip_size.depth_or_array_layers
+        } else {
+            1
+        });
+
+        for array_layer in layer_range.clone() {
+            // TODO: Only doing one layer at a time for volume textures right now.
+            for z in z_range.clone() {
+                // May need multiple copies for each subresource! However, we assume that we never need to split a row.
+                let mut num_rows_left = mip_size.height;
+                while num_rows_left > 0 {
+                    let num_rows = num_rows_left.min(max_rows_per_copy);
+
+                    out_copy_regions.push(hal::BufferTextureCopy {
+                        buffer_layout: wgt::ImageDataLayout {
+                            offset: 0,
+                            bytes_per_row: NonZeroU32::new(bytes_per_row),
+                            rows_per_image: None,
+                        },
+                        texture_base: hal::TextureCopyBase {
+                            mip_level,
+                            array_layer,
+                            origin: wgt::Origin3d {
+                                x: 0, // Always full rows
+                                y: mip_size.height - num_rows_left,
+                                z,
+                            },
+                            aspect: hal::FormatAspects::all(),
+                        },
+                        size: hal::CopyExtent {
+                            width: mip_size.width, // full row
+                            height: num_rows,
+                            depth: 1, // Only single slice of volume texture at a time right now
+                        },
+                    });
+
+                    num_rows_left -= num_rows;
+                }
+            }
+        }
+    }
+}
