@@ -1,19 +1,21 @@
 use crate::{
     binding_model::BindError,
     command::{
-        bind::Binder, end_pipeline_statistics_query, BasePass, BasePassRef, CommandBuffer,
-        CommandEncoderError, CommandEncoderStatus, DrawError, ExecutionError, MapPassErr,
-        PassErrorScope, QueryResetMap, QueryUseError, RenderCommand, RenderCommandError,
-        StateChange,
+        bind::Binder,
+        end_pipeline_statistics_query,
+        memory_init::{fixup_discarded_surfaces, SurfacesInDiscardState},
+        BasePass, BasePassRef, CommandBuffer, CommandEncoderError, CommandEncoderStatus, DrawError,
+        ExecutionError, MapPassErr, PassErrorScope, QueryResetMap, QueryUseError, RenderCommand,
+        RenderCommandError, StateChange,
     },
     device::{
-        AttachmentData, MissingDownlevelFlags, MissingFeatures, RenderPassCompatibilityError,
-        RenderPassContext,
+        AttachmentData, Device, MissingDownlevelFlags, MissingFeatures,
+        RenderPassCompatibilityError, RenderPassContext,
     },
     error::{ErrorFormatter, PrettyError},
     hub::{Global, GlobalIdentityHandlerFactory, HalApi, Storage, Token},
     id,
-    init_tracker::MemoryInitKind,
+    init_tracker::{MemoryInitKind, TextureInitRange, TextureInitTrackerAction},
     pipeline::PipelineFlags,
     resource::{Texture, TextureView},
     track::{StatefulTrackerSubset, TextureSelector, UsageConflict},
@@ -27,7 +29,8 @@ use arrayvec::ArrayVec;
 use hal::CommandEncoder as _;
 use thiserror::Error;
 use wgt::{
-    BufferAddress, BufferSize, BufferUsages, Color, IndexFormat, TextureUsages, VertexStepMode,
+    BufferAddress, BufferSize, BufferUsages, Color, IndexFormat, TextureUsages,
+    TextureViewDimension, VertexStepMode,
 };
 
 #[cfg(any(feature = "serial-pass", feature = "replay"))]
@@ -38,11 +41,14 @@ use serde::Serialize;
 use crate::track::UseExtendError;
 use std::{borrow::Cow, fmt, iter, marker::PhantomData, mem, num::NonZeroU32, ops::Range, str};
 
+use super::{memory_init::TextureSurfaceDiscard, CommandBufferTextureMemoryActions};
+
 /// Operation to perform to the output attachment at the start of a renderpass.
 #[repr(C)]
 #[derive(Copy, Clone, Debug, Hash, Eq, PartialEq)]
 #[cfg_attr(any(feature = "serial-pass", feature = "trace"), derive(Serialize))]
 #[cfg_attr(any(feature = "serial-pass", feature = "replay"), derive(Deserialize))]
+#[cfg_attr(feature = "serde", serde(rename_all = "kebab-case"))]
 pub enum LoadOp {
     /// Clear the output attachment with the clear color. Clearing is faster than loading.
     Clear = 0,
@@ -74,7 +80,7 @@ pub struct PassChannel<V> {
     pub load_op: LoadOp,
     /// Operation to perform to the output attachment at the end of a renderpass.
     pub store_op: StoreOp,
-    /// If load_op is [`LoadOp::Clear`], the attachement will be cleared to this color.
+    /// If load_op is [`LoadOp::Clear`], the attachment will be cleared to this color.
     pub clear_value: V,
     /// If true, the relevant channel is not changed by a renderpass, and the corresponding attachment
     /// can be used inside the pass by other read-only usages.
@@ -340,11 +346,11 @@ struct State {
 impl State {
     fn is_ready(&self, indexed: bool) -> Result<(), DrawError> {
         // Determine how many vertex buffers have already been bound
-        let bound_buffers = self.vertex.inputs.iter().take_while(|v| v.bound).count() as u32;
+        let vertex_buffer_count = self.vertex.inputs.iter().take_while(|v| v.bound).count() as u32;
         // Compare with the needed quantity
-        if bound_buffers < self.vertex.buffers_required {
+        if vertex_buffer_count < self.vertex.buffers_required {
             return Err(DrawError::MissingVertexBuffer {
-                index: bound_buffers,
+                index: vertex_buffer_count,
             });
         }
 
@@ -361,6 +367,7 @@ impl State {
         if self.blend_constant == OptionalState::Required {
             return Err(DrawError::MissingBlendConstant);
         }
+
         if indexed {
             // Pipeline expects an index buffer
             if let Some(pipeline_index_format) = self.index.pipeline_format {
@@ -376,6 +383,9 @@ impl State {
                 }
             }
         }
+
+        self.binder.check_late_buffer_bindings()?;
+
         Ok(())
     }
 
@@ -457,6 +467,12 @@ pub enum RenderPassErrorInner {
     Bind(#[from] BindError),
     #[error(transparent)]
     QueryUse(#[from] QueryUseError),
+    #[error("multiview layer count must match")]
+    MultiViewMismatch,
+    #[error(
+        "multiview pass texture views with more than one array layer must have D2Array dimension"
+    )]
+    MultiViewDimensionMismatch,
 }
 
 impl PrettyError for RenderPassErrorInner {
@@ -531,19 +547,61 @@ type AttachmentDataVec<T> = ArrayVec<T, MAX_TOTAL_ATTACHMENTS>;
 struct RenderPassInfo<'a, A: hal::Api> {
     context: RenderPassContext,
     trackers: StatefulTrackerSubset,
-    render_attachments: AttachmentDataVec<RenderAttachment<'a>>,
+    render_attachments: AttachmentDataVec<RenderAttachment<'a>>, // All render attachments, including depth/stencil
     is_ds_read_only: bool,
     extent: wgt::Extent3d,
     _phantom: PhantomData<A>,
+
+    pending_discard_init_fixups: SurfacesInDiscardState,
+    divergent_discarded_depth_stencil_aspect: Option<(wgt::TextureAspect, &'a TextureView<A>)>,
+    multiview: Option<NonZeroU32>,
 }
 
 impl<'a, A: HalApi> RenderPassInfo<'a, A> {
+    fn add_pass_texture_init_actions<V>(
+        channel: &PassChannel<V>,
+        texture_memory_actions: &mut CommandBufferTextureMemoryActions,
+        view: &TextureView<A>,
+        texture_guard: &Storage<Texture<A>, id::TextureId>,
+        pending_discard_init_fixups: &mut SurfacesInDiscardState,
+    ) {
+        if channel.load_op == LoadOp::Load {
+            pending_discard_init_fixups.extend(texture_memory_actions.register_init_action(
+                &TextureInitTrackerAction {
+                    id: view.parent_id.value.0,
+                    range: TextureInitRange::from(view.selector.clone()),
+                    // Note that this is needed even if the target is discarded,
+                    kind: MemoryInitKind::NeedsInitializedMemory,
+                },
+                texture_guard,
+            ));
+        } else if channel.store_op == StoreOp::Store {
+            // Clear + Store
+            texture_memory_actions.register_implicit_init(
+                view.parent_id.value.0,
+                TextureInitRange::from(view.selector.clone()),
+                texture_guard,
+            );
+        }
+        if channel.store_op == StoreOp::Discard {
+            // the discard happens at the *end* of a pass
+            // but recording the discard right away be alright since the texture can't be used during the pass anyways
+            texture_memory_actions.discard(TextureSurfaceDiscard {
+                texture: view.parent_id.value.0,
+                mip_level: view.selector.levels.start,
+                layer: view.selector.layers.start,
+            });
+        }
+    }
+
     fn start(
+        device: &Device<A>,
         label: Option<&str>,
         color_attachments: &[RenderPassColorAttachment],
         depth_stencil_attachment: Option<&RenderPassDepthStencilAttachment>,
         cmd_buf: &mut CommandBuffer<A>,
         view_guard: &'a Storage<TextureView<A>, id::TextureViewId>,
+        texture_guard: &'a Storage<Texture<A>, id::TextureId>,
     ) -> Result<Self, RenderPassErrorInner> {
         profiling::scope!("start", "RenderPassInfo");
 
@@ -553,11 +611,47 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
         let mut is_ds_read_only = false;
 
         let mut render_attachments = AttachmentDataVec::<RenderAttachment>::new();
+        let mut discarded_surfaces = AttachmentDataVec::new();
+        let mut pending_discard_init_fixups = SurfacesInDiscardState::new();
+        let mut divergent_discarded_depth_stencil_aspect = None;
 
         let mut attachment_type_name = "";
         let mut extent = None;
         let mut sample_count = 0;
 
+        let mut detected_multiview: Option<Option<NonZeroU32>> = None;
+
+        let mut check_multiview = |view: &TextureView<A>| {
+            // Get the multiview configuration for this texture view
+            let layers = view.selector.layers.end - view.selector.layers.start;
+            let this_multiview = if layers >= 2 {
+                // Trivially proven by the if above
+                Some(unsafe { NonZeroU32::new_unchecked(layers) })
+            } else {
+                None
+            };
+
+            // Make sure that if this view is a multiview, it is set to be an array
+            if this_multiview.is_some() && view.desc.dimension != TextureViewDimension::D2Array {
+                return Err(RenderPassErrorInner::MultiViewDimensionMismatch);
+            }
+
+            // Validate matching first, or store the first one
+            if let Some(multiview) = detected_multiview {
+                if multiview != this_multiview {
+                    return Err(RenderPassErrorInner::MultiViewMismatch);
+                }
+            } else {
+                // Multiview is only supported if the feature is enabled
+                if this_multiview.is_some() {
+                    device.require_features(wgt::Features::MULTIVIEW)?;
+                }
+
+                detected_multiview = Some(this_multiview);
+            }
+
+            Ok(())
+        };
         let mut add_view = |view: &TextureView<A>, type_name| {
             if let Some(ex) = extent {
                 if ex != view.extent {
@@ -590,6 +684,7 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
                 .views
                 .use_extend(&*view_guard, at.view, (), ())
                 .map_err(|_| RenderPassErrorInner::InvalidAttachment(at.view))?;
+            check_multiview(view)?;
             add_view(view, "depth")?;
 
             let ds_aspects = view.desc.aspects();
@@ -599,6 +694,80 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
                 ));
             }
 
+            if !ds_aspects.contains(hal::FormatAspects::STENCIL)
+                || (at.stencil.load_op == at.depth.load_op
+                    && at.stencil.store_op == at.depth.store_op)
+            {
+                Self::add_pass_texture_init_actions(
+                    &at.depth,
+                    &mut cmd_buf.texture_memory_actions,
+                    view,
+                    texture_guard,
+                    &mut pending_discard_init_fixups,
+                );
+            } else if !ds_aspects.contains(hal::FormatAspects::DEPTH) {
+                Self::add_pass_texture_init_actions(
+                    &at.stencil,
+                    &mut cmd_buf.texture_memory_actions,
+                    view,
+                    texture_guard,
+                    &mut pending_discard_init_fixups,
+                );
+            } else {
+                // This is the only place (anywhere in wgpu) where Stencil & Depth init state can diverge.
+                // To safe us the overhead of tracking init state of texture aspects everywhere,
+                // we're going to cheat a little bit in order to keep the init state of both Stencil and Depth aspects in sync.
+                // The expectation is that we hit this path extremely rarely!
+
+                // Diverging LoadOp, i.e. Load + Clear:
+                // Record MemoryInitKind::NeedsInitializedMemory for the entire surface, a bit wasteful on unit but no negative effect!
+                // Rationale: If the loaded channel is uninitialized it needs clearing, the cleared channel doesn't care. (If everything is already initialized nothing special happens)
+                // (possible minor optimization: Clear caused by NeedsInitializedMemory should know that it doesn't need to clear the aspect that was set to C)
+                let need_init_beforehand =
+                    at.depth.load_op == LoadOp::Load || at.stencil.load_op == LoadOp::Load;
+                if need_init_beforehand {
+                    pending_discard_init_fixups.extend(
+                        cmd_buf.texture_memory_actions.register_init_action(
+                            &TextureInitTrackerAction {
+                                id: view.parent_id.value.0,
+                                range: TextureInitRange::from(view.selector.clone()),
+                                kind: MemoryInitKind::NeedsInitializedMemory,
+                            },
+                            texture_guard,
+                        ),
+                    );
+                }
+
+                // Diverging Store, i.e. Discard + Store:
+                // Immediately zero out channel that is set to discard after we're done with the render pass.
+                // This allows us to set the entire surface to MemoryInitKind::ImplicitlyInitialized (if it isn't already set to NeedsInitializedMemory).
+                // (possible optimization: Delay and potentially drop this zeroing)
+                if at.depth.store_op != at.stencil.store_op {
+                    if !need_init_beforehand {
+                        cmd_buf.texture_memory_actions.register_implicit_init(
+                            view.parent_id.value.0,
+                            TextureInitRange::from(view.selector.clone()),
+                            texture_guard,
+                        );
+                    }
+                    divergent_discarded_depth_stencil_aspect = Some((
+                        if at.depth.store_op == StoreOp::Discard {
+                            wgt::TextureAspect::DepthOnly
+                        } else {
+                            wgt::TextureAspect::StencilOnly
+                        },
+                        view,
+                    ));
+                } else if at.depth.store_op == StoreOp::Discard {
+                    // Both are discarded using the regular path.
+                    discarded_surfaces.push(TextureSurfaceDiscard {
+                        texture: view.parent_id.value.0,
+                        mip_level: view.selector.levels.start,
+                        layer: view.selector.layers.start,
+                    });
+                }
+            }
+
             let usage = if at.is_read_only(ds_aspects)? {
                 is_ds_read_only = true;
                 hal::TextureUses::DEPTH_STENCIL_READ | hal::TextureUses::RESOURCE
@@ -624,6 +793,7 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
                 .views
                 .use_extend(&*view_guard, at.view, (), ())
                 .map_err(|_| RenderPassErrorInner::InvalidAttachment(at.view))?;
+            check_multiview(color_view)?;
             add_view(color_view, "color")?;
 
             if !color_view
@@ -636,6 +806,13 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
                 ));
             }
 
+            Self::add_pass_texture_init_actions(
+                &at.channel,
+                &mut cmd_buf.texture_memory_actions,
+                color_view,
+                texture_guard,
+                &mut pending_discard_init_fixups,
+            );
             render_attachments
                 .push(color_view.to_render_attachment(hal::TextureUses::COLOR_TARGET));
 
@@ -646,6 +823,7 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
                     .views
                     .use_extend(&*view_guard, resolve_target, (), ())
                     .map_err(|_| RenderPassErrorInner::InvalidAttachment(resolve_target))?;
+                check_multiview(resolve_view)?;
                 if color_view.extent != resolve_view.extent {
                     return Err(RenderPassErrorInner::AttachmentsDimensionMismatch {
                         previous: (attachment_type_name, extent.unwrap_or_default()),
@@ -659,6 +837,11 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
                     return Err(RenderPassErrorInner::InvalidResolveTargetSampleCount);
                 }
 
+                cmd_buf.texture_memory_actions.register_implicit_init(
+                    resolve_view.parent_id.value.0,
+                    TextureInitRange::from(resolve_view.selector.clone()),
+                    texture_guard,
+                );
                 render_attachments
                     .push(resolve_view.to_render_attachment(hal::TextureUses::COLOR_TARGET));
 
@@ -696,9 +879,12 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
             depth_stencil: depth_stencil_attachment.map(|at| view_guard.get(at.view).unwrap()),
         };
         let extent = extent.ok_or(RenderPassErrorInner::MissingAttachments)?;
+
+        let multiview = detected_multiview.expect("Multiview was not detected, no attachments");
         let context = RenderPassContext {
             attachments: view_data.map(|view| view.desc.format),
             sample_count,
+            multiview,
         };
 
         let hal_desc = hal::RenderPassDescriptor {
@@ -707,6 +893,7 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
             sample_count,
             color_attachments: &colors,
             depth_stencil_attachment: depth_stencil,
+            multiview,
         };
         unsafe {
             cmd_buf.encoder.raw.begin_render_pass(&hal_desc);
@@ -719,6 +906,9 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
             is_ds_read_only,
             extent,
             _phantom: PhantomData,
+            pending_discard_init_fixups,
+            divergent_discarded_depth_stencil_aspect,
+            multiview,
         })
     }
 
@@ -726,7 +916,7 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
         mut self,
         raw: &mut A::CommandEncoder,
         texture_guard: &Storage<Texture<A>, id::TextureId>,
-    ) -> Result<StatefulTrackerSubset, RenderPassErrorInner> {
+    ) -> Result<(StatefulTrackerSubset, SurfacesInDiscardState), RenderPassErrorInner> {
         profiling::scope!("finish", "RenderPassInfo");
         unsafe {
             raw.end_render_pass();
@@ -751,7 +941,45 @@ impl<'a, A: HalApi> RenderPassInfo<'a, A> {
                 .map_err(UsageConflict::from)?;
         }
 
-        Ok(self.trackers)
+        // If either only stencil or depth was discarded, we put in a special clear pass to keep the init status of the aspects in sync.
+        // We do this so we don't need to track init state for depth/stencil aspects individually.
+        // Note that we don't go the usual route of "brute force" initializing the texture when need arises here,
+        // since this path is actually something a user may genuinely want (where as the other cases are more seen along the lines as gracefully handling a user error).
+        if let Some((aspect, view)) = self.divergent_discarded_depth_stencil_aspect {
+            let (depth_ops, stencil_ops) = if aspect == wgt::TextureAspect::DepthOnly {
+                (
+                    hal::AttachmentOps::STORE,                            // clear depth
+                    hal::AttachmentOps::LOAD | hal::AttachmentOps::STORE, // unchanged stencil
+                )
+            } else {
+                (
+                    hal::AttachmentOps::LOAD | hal::AttachmentOps::STORE, // unchanged stencil
+                    hal::AttachmentOps::STORE,                            // clear depth
+                )
+            };
+            let desc = hal::RenderPassDescriptor {
+                label: Some("Zero init discarded depth/stencil aspect"),
+                extent: view.extent,
+                sample_count: view.samples,
+                color_attachments: &[],
+                depth_stencil_attachment: Some(hal::DepthStencilAttachment {
+                    target: hal::Attachment {
+                        view: &view.raw,
+                        usage: hal::TextureUses::DEPTH_STENCIL_WRITE,
+                    },
+                    depth_ops,
+                    stencil_ops,
+                    clear_value: (0.0, 0),
+                }),
+                multiview: self.multiview,
+            };
+            unsafe {
+                raw.begin_render_pass(&desc);
+                raw.end_render_pass();
+            }
+        }
+
+        Ok((self.trackers, self.pending_discard_init_fixups))
     }
 }
 
@@ -780,17 +1008,17 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         depth_stencil_attachment: Option<&RenderPassDepthStencilAttachment>,
     ) -> Result<(), RenderPassError> {
         profiling::scope!("run_render_pass", "CommandEncoder");
-        let scope = PassErrorScope::Pass(encoder_id);
+        let init_scope = PassErrorScope::Pass(encoder_id);
 
         let hub = A::hub(self);
         let mut token = Token::root();
         let (device_guard, mut token) = hub.devices.read(&mut token);
 
-        let (pass_raw, trackers, query_reset_state) = {
+        let (trackers, query_reset_state, pending_discard_init_fixups) = {
             let (mut cmb_guard, mut token) = hub.command_buffers.write(&mut token);
 
-            let cmd_buf =
-                CommandBuffer::get_encoder_mut(&mut *cmb_guard, encoder_id).map_pass_err(scope)?;
+            let cmd_buf = CommandBuffer::get_encoder_mut(&mut *cmb_guard, encoder_id)
+                .map_pass_err(init_scope)?;
             // close everything while the new command encoder is filled
             cmd_buf.encoder.close();
             // will be reset to true if recording is done without errors
@@ -806,9 +1034,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
             }
 
             let device = &device_guard[cmd_buf.device_id.value];
-            unsafe {
-                cmd_buf.encoder.raw.begin_encoding(base.label).unwrap() //TODO: handle this better
-            };
+            cmd_buf.encoder.open_pass(base.label);
 
             let (bundle_guard, mut token) = hub.render_bundles.read(&mut token);
             let (pipeline_layout_guard, mut token) = hub.pipeline_layouts.read(&mut token);
@@ -825,13 +1051,15 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
             );
 
             let mut info = RenderPassInfo::start(
+                device,
                 base.label,
                 color_attachments,
                 depth_stencil_attachment,
                 cmd_buf,
                 &*view_guard,
+                &*texture_guard,
             )
-            .map_pass_err(scope)?;
+            .map_pass_err(init_scope)?;
 
             let raw = &mut cmd_buf.encoder.raw;
 
@@ -900,6 +1128,13 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                                 }
                             }),
                         );
+                        for action in bind_group.used_texture_ranges.iter() {
+                            info.pending_discard_init_fixups.extend(
+                                cmd_buf
+                                    .texture_memory_actions
+                                    .register_init_action(action, &texture_guard),
+                            );
+                        }
 
                         let pipeline_layout_id = state.binder.pipeline_layout_id;
                         let entries = state.binder.assign_group(
@@ -974,6 +1209,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                             let (start_index, entries) = state.binder.change_pipeline_layout(
                                 &*pipeline_layout_guard,
                                 pipeline.layout_id.value,
+                                &pipeline.late_sized_buffer_groups,
                             );
                             if !entries.is_empty() {
                                 for (i, e) in entries.iter().enumerate() {
@@ -1630,6 +1866,13 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                                     Err(_) => None,
                                 }),
                         );
+                        for action in bundle.texture_memory_init_actions.iter() {
+                            info.pending_discard_init_fixups.extend(
+                                cmd_buf
+                                    .texture_memory_actions
+                                    .register_init_action(action, &texture_guard),
+                            );
+                        }
 
                         unsafe {
                             bundle.execute(
@@ -1667,15 +1910,11 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
             }
 
             log::trace!("Merging {:?} with the render pass", encoder_id);
-            let trackers = info.finish(raw, &*texture_guard).map_pass_err(scope)?;
+            let (trackers, pending_discard_init_fixups) =
+                info.finish(raw, &*texture_guard).map_pass_err(init_scope)?;
 
-            let raw_cmd_buf = unsafe {
-                raw.end_encoding()
-                    .map_err(|_| RenderPassErrorInner::OutOfMemory)
-                    .map_pass_err(scope)?
-            };
-            cmd_buf.status = CommandEncoderStatus::Recording;
-            (raw_cmd_buf, trackers, query_reset_state)
+            cmd_buf.encoder.close();
+            (trackers, query_reset_state, pending_discard_init_fixups)
         };
 
         let (mut cmb_guard, mut token) = hub.command_buffers.write(&mut token);
@@ -1683,10 +1922,18 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         let (buffer_guard, mut token) = hub.buffers.read(&mut token);
         let (texture_guard, _) = hub.textures.read(&mut token);
 
-        let cmd_buf =
-            CommandBuffer::get_encoder_mut(&mut *cmb_guard, encoder_id).map_pass_err(scope)?;
+        let cmd_buf = cmb_guard.get_mut(encoder_id).unwrap();
         {
             let transit = cmd_buf.encoder.open();
+
+            fixup_discarded_surfaces(
+                pending_discard_init_fixups.into_iter(),
+                transit,
+                &texture_guard,
+                &mut cmd_buf.trackers.textures,
+                &device_guard[cmd_buf.device_id.value],
+            );
+
             query_reset_state
                 .reset_queries(
                     transit,
@@ -1705,8 +1952,16 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                 &*texture_guard,
             );
         }
+
+        // Before we finish the auxiliary encoder, let's
+        // get our pass back and place it after.
+        //Note: we could just hold onto this raw pass while recording the
+        // auxiliary encoder, but then handling errors and cleaning up
+        // would be more complicated, so we re-use `open()`/`close()`.
+        let pass_raw = cmd_buf.encoder.list.pop().unwrap();
         cmd_buf.encoder.close();
         cmd_buf.encoder.list.push(pass_raw);
+        cmd_buf.status = CommandEncoderStatus::Recording;
 
         Ok(())
     }
