@@ -1,12 +1,16 @@
 #[cfg(feature = "trace")]
 use crate::device::trace::Command as TraceCommand;
 use crate::{
-    command::{CommandBuffer, CommandEncoderError},
+    command::{
+        collect_zero_buffer_copies_for_clear_texture, memory_init::fixup_discarded_surfaces,
+        CommandBuffer, CommandEncoderError,
+    },
     conv,
+    device::Device,
     error::{ErrorFormatter, PrettyError},
     hub::{Global, GlobalIdentityHandlerFactory, HalApi, Storage, Token},
     id::{BufferId, CommandEncoderId, TextureId},
-    init_tracker::MemoryInitKind,
+    init_tracker::{MemoryInitKind, TextureInitRange, TextureInitTrackerAction},
     resource::{Texture, TextureErrorDimension},
     track::TextureSelector,
 };
@@ -91,6 +95,15 @@ pub enum TransferError {
     CopyFromForbiddenTextureFormat(wgt::TextureFormat),
     #[error("copying to textures with format {0:?} is forbidden")]
     CopyToForbiddenTextureFormat(wgt::TextureFormat),
+    #[error("the entire texture must be copied when copying from depth texture")]
+    InvalidDepthTextureExtent,
+    #[error(
+        "source format ({src_format:?}) and destination format ({dst_format:?}) are different"
+    )]
+    MismatchedTextureFormats {
+        src_format: wgt::TextureFormat,
+        dst_format: wgt::TextureFormat,
+    },
 }
 
 impl PrettyError for TransferError {
@@ -291,6 +304,17 @@ pub(crate) fn validate_texture_copy_range(
     // physical size can be larger than the virtual
     let extent = extent_virtual.physical_size(desc.format);
 
+    match desc.format {
+        wgt::TextureFormat::Depth32Float
+        | wgt::TextureFormat::Depth24Plus
+        | wgt::TextureFormat::Depth24PlusStencil8 => {
+            if *copy_size != extent {
+                return Err(TransferError::InvalidDepthTextureExtent);
+            }
+        }
+        _ => {}
+    }
+
     let x_copy_max = texture_copy_view.origin.x + copy_size.width;
     if x_copy_max > extent.width {
         return Err(TransferError::TextureOverrun {
@@ -350,6 +374,62 @@ pub(crate) fn validate_texture_copy_range(
     Ok((copy_extent, array_layer_count))
 }
 
+fn get_copy_dst_texture_init_requirement<A: HalApi>(
+    texture: &Texture<A>,
+    copy_texture: &wgt::ImageCopyTexture<TextureId>,
+    copy_size: &Extent3d,
+) -> TextureInitTrackerAction {
+    // Attention: If we don't write full texture subresources, we need to a full clear first since we don't track subrects.
+    let dst_init_kind = if copy_size.width == texture.desc.size.width
+        && copy_size.height == texture.desc.size.height
+    {
+        MemoryInitKind::ImplicitlyInitialized
+    } else {
+        MemoryInitKind::NeedsInitializedMemory
+    };
+    TextureInitTrackerAction {
+        id: copy_texture.texture,
+        range: TextureInitRange {
+            mip_range: copy_texture.mip_level..copy_texture.mip_level + 1,
+            layer_range: copy_texture.origin.z
+                ..(copy_texture.origin.z + copy_size.depth_or_array_layers),
+        },
+        kind: dst_init_kind,
+    }
+}
+
+fn handle_src_texture_init<A: hal::Api>(
+    cmd_buf: &mut CommandBuffer<A>,
+    device: &Device<A>,
+    source: &ImageCopyTexture,
+    src_base: &hal::TextureCopyBase,
+    copy_size: &Extent3d,
+    texture_guard: &Storage<Texture<A>, TextureId>,
+) {
+    let immediate_src_init = cmd_buf.texture_memory_actions.register_init_action(
+        &TextureInitTrackerAction {
+            id: source.texture,
+            range: TextureInitRange {
+                mip_range: src_base.mip_level..src_base.mip_level + 1,
+                layer_range: src_base.origin.z
+                    ..(src_base.origin.z + copy_size.depth_or_array_layers),
+            },
+            kind: MemoryInitKind::NeedsInitializedMemory,
+        },
+        texture_guard,
+    );
+    if !immediate_src_init.is_empty() {
+        let cmd_buf_raw = cmd_buf.encoder.open();
+        fixup_discarded_surfaces(
+            immediate_src_init.into_iter(),
+            cmd_buf_raw,
+            texture_guard,
+            &mut cmd_buf.trackers.textures,
+            device,
+        );
+    }
+}
+
 impl<G: GlobalIdentityHandlerFactory> Global<G> {
     pub fn command_encoder_copy_buffer_to_buffer<A: HalApi>(
         &self,
@@ -493,11 +573,14 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         let hub = A::hub(self);
         let mut token = Token::root();
 
+        let (device_guard, mut token) = hub.devices.read(&mut token);
         let (mut cmd_buf_guard, mut token) = hub.command_buffers.write(&mut token);
         let cmd_buf = CommandBuffer::get_encoder_mut(&mut *cmd_buf_guard, command_encoder_id)?;
         let (buffer_guard, mut token) = hub.buffers.read(&mut token);
         let (texture_guard, _) = hub.textures.read(&mut token);
 
+        let device = &device_guard[cmd_buf.device_id.value];
+
         #[cfg(feature = "trace")]
         if let Some(ref mut list) = cmd_buf.commands {
             list.push(TraceCommand::CopyBufferToTexture {
@@ -567,6 +650,12 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
             true,
         )?;
 
+        if !conv::is_valid_copy_dst_texture_format(dst_texture.desc.format) {
+            return Err(
+                TransferError::CopyToForbiddenTextureFormat(dst_texture.desc.format).into(),
+            );
+        }
+
         cmd_buf
             .buffer_memory_init_actions
             .extend(src_buffer.initialization_status.create_action(
@@ -574,10 +663,17 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                 source.layout.offset..(source.layout.offset + required_buffer_bytes_in_copy),
                 MemoryInitKind::NeedsInitializedMemory,
             ));
-
-        if !conv::is_valid_copy_dst_texture_format(dst_texture.desc.format) {
-            return Err(
-                TransferError::CopyToForbiddenTextureFormat(dst_texture.desc.format).into(),
+        let mut dst_zero_buffer_copy_regions = Vec::new();
+        for immediate_init in cmd_buf.texture_memory_actions.register_init_action(
+            &get_copy_dst_texture_init_requirement(dst_texture, destination, copy_size),
+            &texture_guard,
+        ) {
+            collect_zero_buffer_copies_for_clear_texture(
+                &dst_texture.desc,
+                device.alignments.buffer_copy_pitch.get() as u32,
+                immediate_init.mip_level..(immediate_init.mip_level + 1),
+                immediate_init.layer..(immediate_init.layer + 1),
+                &mut dst_zero_buffer_copy_regions,
             );
         }
 
@@ -594,8 +690,16 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         });
         let cmd_buf_raw = cmd_buf.encoder.open();
         unsafe {
-            cmd_buf_raw.transition_buffers(src_barriers);
             cmd_buf_raw.transition_textures(dst_barriers);
+            // potential dst buffer init (for previously discarded dst_texture + partial copy)
+            if !dst_zero_buffer_copy_regions.is_empty() {
+                cmd_buf_raw.copy_buffer_to_texture(
+                    &device.zero_buffer,
+                    dst_raw,
+                    dst_zero_buffer_copy_regions.into_iter(),
+                );
+            }
+            cmd_buf_raw.transition_buffers(src_barriers);
             cmd_buf_raw.copy_buffer_to_texture(src_raw, dst_raw, regions);
         }
         Ok(())
@@ -613,11 +717,14 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         let hub = A::hub(self);
         let mut token = Token::root();
 
+        let (device_guard, mut token) = hub.devices.read(&mut token);
         let (mut cmd_buf_guard, mut token) = hub.command_buffers.write(&mut token);
         let cmd_buf = CommandBuffer::get_encoder_mut(&mut *cmd_buf_guard, command_encoder_id)?;
         let (buffer_guard, mut token) = hub.buffers.read(&mut token);
         let (texture_guard, _) = hub.textures.read(&mut token);
 
+        let device = &device_guard[cmd_buf.device_id.value];
+
         #[cfg(feature = "trace")]
         if let Some(ref mut list) = cmd_buf.commands {
             list.push(TraceCommand::CopyTextureToBuffer {
@@ -635,6 +742,16 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         let (src_range, src_base, _) =
             extract_texture_selector(source, copy_size, &*texture_guard)?;
 
+        // Handle src texture init *before* dealing with barrier transitions so we have an easier time inserting "immediate-inits" that may be required by prior discards in rare cases.
+        handle_src_texture_init(
+            cmd_buf,
+            device,
+            source,
+            &src_base,
+            copy_size,
+            &texture_guard,
+        );
+
         let (src_texture, src_pending) = cmd_buf
             .trackers
             .textures
@@ -740,11 +857,14 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         let hub = A::hub(self);
         let mut token = Token::root();
 
+        let (device_guard, mut token) = hub.devices.read(&mut token);
         let (mut cmd_buf_guard, mut token) = hub.command_buffers.write(&mut token);
         let cmd_buf = CommandBuffer::get_encoder_mut(&mut *cmd_buf_guard, command_encoder_id)?;
         let (_, mut token) = hub.buffers.read(&mut token); // skip token
         let (texture_guard, _) = hub.textures.read(&mut token);
 
+        let device = &device_guard[cmd_buf.device_id.value];
+
         #[cfg(feature = "trace")]
         if let Some(ref mut list) = cmd_buf.commands {
             list.push(TraceCommand::CopyTextureToTexture {
@@ -767,6 +887,16 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
             return Err(TransferError::MismatchedAspects.into());
         }
 
+        // Handle src texture init *before* dealing with barrier transitions so we have an easier time inserting "immediate-inits" that may be required by prior discards in rare cases.
+        handle_src_texture_init(
+            cmd_buf,
+            device,
+            source,
+            &src_tex_base,
+            copy_size,
+            &texture_guard,
+        );
+
         let (src_texture, src_pending) = cmd_buf
             .trackers
             .textures
@@ -784,6 +914,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         if !src_texture.desc.usage.contains(TextureUsages::COPY_SRC) {
             return Err(TransferError::MissingCopySrcUsageFlag.into());
         }
+
         //TODO: try to avoid this the collection. It's needed because both
         // `src_pending` and `dst_pending` try to hold `trackers.textures` mutably.
         let mut barriers = src_pending
@@ -809,6 +940,19 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                 TransferError::MissingCopyDstUsageFlag(None, Some(destination.texture)).into(),
             );
         }
+
+        // src and dst texture format must be the same.
+        let src_format = src_texture.desc.format;
+        let dst_format = dst_texture.desc.format;
+
+        if src_format != dst_format {
+            return Err(TransferError::MismatchedTextureFormats {
+                src_format,
+                dst_format,
+            }
+            .into());
+        }
+
         barriers.extend(dst_pending.map(|pending| pending.into_hal(dst_texture)));
 
         let (src_copy_size, array_layer_count) =
@@ -820,6 +964,20 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
             copy_size,
         )?;
 
+        let mut dst_zero_buffer_copy_regions = Vec::new();
+        for immediate_init in cmd_buf.texture_memory_actions.register_init_action(
+            &get_copy_dst_texture_init_requirement(dst_texture, destination, copy_size),
+            &texture_guard,
+        ) {
+            collect_zero_buffer_copies_for_clear_texture(
+                &dst_texture.desc,
+                device.alignments.buffer_copy_pitch.get() as u32,
+                immediate_init.mip_level..(immediate_init.mip_level + 1),
+                immediate_init.layer..(immediate_init.layer + 1),
+                &mut dst_zero_buffer_copy_regions,
+            );
+        }
+
         let hal_copy_size = hal::CopyExtent {
             width: src_copy_size.width.min(dst_copy_size.width),
             height: src_copy_size.height.min(dst_copy_size.height),
@@ -839,6 +997,16 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         let cmd_buf_raw = cmd_buf.encoder.open();
         unsafe {
             cmd_buf_raw.transition_textures(barriers.into_iter());
+
+            // potential dst buffer init (for previously discarded dst_texture + partial copy)
+            if !dst_zero_buffer_copy_regions.is_empty() {
+                cmd_buf_raw.copy_buffer_to_texture(
+                    &device.zero_buffer,
+                    dst_raw,
+                    dst_zero_buffer_copy_regions.into_iter(),
+                );
+            }
+
             cmd_buf_raw.copy_texture_to_texture(
                 src_raw,
                 hal::TextureUses::COPY_SRC,
